#!c:/usr/bin/perl -w

# M3U Playlist Director
# This Perl script generates extended M3U playlists from a given directory.
# Based on extm3u.pl (2011-12-01) by Andreas Gohr <andi@splitbrain.org>

our $version = '2';    # 2023-01-08

use 5.010;
use constant DEBUG => 0;    # Set to 1 to print debug information on the console
use Cwd::utf8 qw( abs_path );
		# Note: Cwd functions are not replaced on DOS, Windows, and OS/2
		# as these systems do not have full UTF-8 file system support.
use if (DEBUG), qw( Data::Dumper );
use Encode qw( decode encode );
use File::Basename;
use File::Spec;
use File::Spec::Functions;
use Getopt::Long;
use I18N::Langinfo qw(langinfo CODESET);
use Module::Load::Conditional qw(check_install);
use Module::Load qw(load);
use open ":std", ":encoding(UTF-8)";
use Perl::OSType qw(is_os_type);
use Scalar::Util qw(reftype);
use strict;
use Term::UI;
use Term::ReadLine;
use vars
  qw( @all $inputpath $LIST_AUDIO $LIST_FLAC $LIST_OGG $LIST_MP3 %OPT $VERSION );
use if ( is_os_type('Windows') ), qw( Win32::LongPath );
use if ( is_os_type('Windows') ), qw( Win32::Unicode );
use warnings 'all';

if ( check_install( module => 'Audio::Scan' ) ) {
    load('Audio::Scan');
    $LIST_AUDIO = 1;
}
else {
    if ( check_install( module => 'MP3::Info' ) ) {
        load MP3::Info;
        $LIST_MP3 = 1;
    }

    if ( check_install( module => 'Ogg::Vorbis::Header::PurePerl' ) ) {
        load('Ogg::Vorbis::Header::PurePerl');
        $LIST_OGG = 1;
    }

    if ( check_install( module => 'Audio::FLAC::Header' ) ) {
        load('Audio::FLAC::Header');
        $LIST_FLAC = 1;
    }
}

# Function prototypes:
sub arrayToString($$);
sub help();
sub printFile($$$$$);
sub readFiles($$);
sub readdirectory($);
sub readdirectoryWindows($);
sub scanAudio($);
sub scanFLAC($);
sub scanMP3($);
sub scanOGG($);
sub validateFilename($);

############################################################################
# Main

# Ensure that possible error messages printed to screen are converted to UTF-8.
# For this to work: Check that you terminal emulator is using UTF-8.
binmode STDOUT, ':utf8';
binmode STDERR, ':utf8';

my %OPT = ();
GetOptions(
    \%OPT, 'album|b', 'all|a', 'disc|d',
    'help|h' => sub { &help() },
    'interactive|i', 'm3udir|m=s', 'random|r', 'twonky|t',
    'version|v' => sub { &version() }
);

my $codeset = langinfo(CODESET)
  ; # Query the character code set being used (such as "ISO8859-1", "cp850", "koi8-r", "sjis", "UTF-8", etc.)
print 'Local character code set: ', $codeset, "\n" if DEBUG;

print '%OPT from the command line: ', "\n", Dumper( \%OPT ) if DEBUG;

if ( $OPT{'m3udir'} ) {
    $OPT{'m3udir'} = decode( $codeset, $OPT{'m3udir'} )
      ;   # Transcode the M3U directory name to the Perl internal utf8 encoding.
    if (DEBUG) {
        if ( is_os_type('Windows') ) {
            printW( "M3U playlists output directory:  ", $OPT{'m3udir'}, "\n" );
        }
        else {
            print( "M3U playlists output directory:  ", $OPT{'m3udir'}, "\n" );
        }
    }
}

# print help page
if ( @ARGV == 0 ) {
    help();
    exit -1;
}

print '@ARGV before transcoding: ', Dumper( \@ARGV ) if DEBUG;
@ARGV = map { decode( $codeset, $_ ) }
  @ARGV;   # Transcode input directory names to the Perl internal utf8 encoding.
print '@ARGV after transcoding: ', Dumper( \@ARGV ) if DEBUG;

if ( $OPT{'interactive'} ) {
    my $term;
    if ( is_os_type('Windows') ) {
        open( my $saveerr, ">&STDERR" );
        close(STDERR);
        $term = Term::ReadLine->new('options')
          ;    # This function causes an error message on Windows OS which
               # is prevented by temporarily closing STDERR.
        open( STDERR, ">&", $saveerr );
    }
    else {
        $term = Term::ReadLine->new('options');
    }
    $OPT{'album'} = $term->ask_yn(
        prompt =>
          'Create playlists for albums that may have more than one artist?',
        default => $OPT{'album'} ? 'y' : 'n'
    );
    $OPT{'all'} = $term->ask_yn(
        prompt  => 'Create playlists with all discs for each album?',
        default => $OPT{'all'} ? 'y' : 'n'
    );
    $OPT{'disc'} = $term->ask_yn(
        prompt =>
          'Create playlists per disc for albums with more than one disc?',
        default => $OPT{'disc'} ? 'y' : 'n'
    );
    $OPT{'m3udir'} = $term->get_reply(
        print_me => 'Shall the playlists be stored in a separate directory?',
        prompt   => 'Then enter this path here',
        default  => $OPT{'m3udir'} ? encode( $codeset, $OPT{'m3udir'} ) : ''
    );
    $OPT{'twonky'} = $term->ask_yn(
        prompt  => 'Will the playlists be used on a Twonky (tm) Server?',
        default => $OPT{'twonky'} ? 'y' : 'n'
    );
    if ( defined $OPT{'m3udir'} && length( $OPT{'m3udir'} ) < 1 ) {
        $OPT{'m3udir'} = undef;
    }
    else {
        $OPT{'m3udir'} = decode( $codeset, $OPT{'m3udir'} )
          ; # Transcode the M3U directory name to the Perl internal utf8 encoding.
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "M3U playlists output directory:  ", $OPT{'m3udir'}, "\n" );
            }
            else {
                print "M3U playlists output directory:  ", $OPT{'m3udir'}, "\n";
            }
        }
    }
    print '%OPT after interactive user input: ', "\n", Dumper( \%OPT ) if DEBUG;
}

my @all;    # This array will contain the audio files found.

while ( $inputpath = shift(@ARGV) ) {

printW($inputpath, "\n");
printW(File::Spec->rel2abs( $inputpath ), "\n");

    my $inputpathabs = abs_path($inputpath);
    if ( !defined $inputpathabs || length($inputpathabs) < 1 ) {
        die "This input path cannot be processed with abs_path(): ",
          $inputpath, "\n";
    }
    if ( is_os_type('Windows') ) {
        printW( '$inputpath: ',    $inputpath,    "\n" ) if DEBUG;
        printW( '$inputpathabs: ', $inputpathabs, "\n" ) if DEBUG;
    }
    else {
        print '$inputpath: ',    $inputpath,    "\n" if DEBUG;
        print '$inputpathabs: ', $inputpathabs, "\n" if DEBUG;
    }
    $inputpath = $inputpathabs;

    # gather or directly print the files
    if ( $OPT{'all'} || $OPT{'disc'} || $OPT{'random'} ) {
        push( @all, readFiles( $inputpath, 1 ) );
        $Data::Dumper::Indent = 3;
        $Data::Dumper::Indent = 3;    # This second setting prevents a warning message.
        print 'All audio files found: ', "\n", Dumper( \@all ) if DEBUG;
    }
    else {
        readFiles( $inputpath, 0 );
    }

    # randomize output to STDOUT
    if ( $OPT{'random'} ) {
        fisher_yates_shuffle( \@all );
        foreach my $file (@all) {
            printFile( undef, $$file[0], $$file[1][0], $$file[1][1],
                $$file[1][4] );
        }
        exit -1;
    }

    # Prepare the gathered albums' and discs' data for later printing
    my %albums;
    my %discs;
    my %discsall;
    my @tracknomissing = ();

    foreach my $file (@all) {
        my $path = ( fileparse( $$file[0] ) )[1];
        if ( !defined $path or length($path) < 1 ) {
            die "The path cannot be determined: ", Dumper( $$file[0] ), "\n";
        }

        my $artist      = $$file[1][1];
        my $album       = $$file[1][2];
        my $albumartist = $$file[1][3];
        my $track       = $$file[1][5];
        my $disc        = $$file[1][6];

        if ( !defined $artist || length($artist) < 1 ) {
            $artist = '';
        }

        if ( !defined $album || length($album) < 1 ) {
            $album = '';
        }

        if ( !defined $albumartist || length($albumartist) < 1 ) {
            $albumartist = '';
        }

        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "File:         ", $$file[0],    "\n" );
                printW( "Path:         ", $path,        "\n" );
                printW( "Artist:       ", $artist,      "\n" );
                printW( "Album:        ", $album,       "\n" );
                printW( "Album artist: ", $albumartist, "\n" );
            }
            else {
                print "File:         ", $$file[0],    "\n";
                print "Path:         ", $path,        "\n";
                print "Artist:       ", $artist,      "\n";
                print "Album:        ", $album,       "\n";
                print "Album artist: ", $albumartist, "\n";
            }
        }

        if ( defined $disc && length($disc) > 0 )
        {    # Does the disc number exist?
            print "Disc (original):   ", $disc, "\n" if DEBUG;
            $disc =~ s/^0*(\d+)[\D]*\d*/$1/
              ; # Disc numbers may look like 03/14, 03 or 3. We need a numeric value.
            if ( $disc =~ m/^\d+$/ ) {
                print "Disc (processed):  ", $disc, "\n" if DEBUG;
            }
            else {
                # The disc number format is invalid.
                $disc = 0;
                print "Disc (processed): invalid format\n" if DEBUG;
            }
        }
        else {
            # the disc number is missing
            $disc = 0;
            print "Disc: 0 (missing)\n" if DEBUG;
        }

        if ( defined $track && length($track) > 0 )
        {    # Does the track number exist?
            print "Track (original):  ", $track, "\n" if DEBUG;
            $track =~ s/^0*(\d+)[\D]*\d*/$1/
              ; # Track numbers may look like 03/14, 03 or 3. We need a numeric value.
            if ( $track =~ m/^\d+$/ ) {
                print "Track (processed): ", $track, "\n" if DEBUG;
            }
            else {
                # The track number format is invalid.
                $track = 0;
                print "Track (processed): invalid format\n" if DEBUG;
            }
        }
        else {
            # the track number is missing
            $track = 0;
            print "Track:\n" if DEBUG;
        }

        if ( $OPT{'album'} ) {

            # An album may have audio files from various artists.
            $discs{$path}{$albumartist}{$album}{$disc}{$track}{ $$file[0] } =
              $file;
        }
        else {
            $discs{$path}{$artist}{$album}{$disc}{$track}{ $$file[0] } = $file;
        }
    }

    if ( $OPT{'m3udir'} ) {
        $OPT{'m3udir'} = abs_path( $OPT{'m3udir'} );
        if ( is_os_type('Windows') ) {
            printW( 'Absolute playlist directory: ', $OPT{'m3udir'}, "\n" )
              if DEBUG;
        }
        else {
            print 'Absolute playlist directory: ', $OPT{'m3udir'}, "\n"
              if DEBUG;
        }
    }

    if ( $OPT{'all'} ) {

        # Create a playlist with all discs of an album per directory
        print
"\n---------------------------------\nGenerating playlists for all artists, discs and tracks of an album:\n\n";
        foreach my $path ( keys %discs ) {
            if (DEBUG) {
                if ( is_os_type('Windows') ) {
                    printW( '$path: ', $path, "\n" );
                }
                else {
                    print '$path: ', $path, "\n";
                }
            }
            if ( $OPT{'album'} ) {

# %discs has the following keys: $discs{$path}{$albumartist}{$album}{$disc}{$track}{$$file[0]}

                foreach my $albumartist ( sort keys %{ $discs{$path} } )
                {    # sort by albumartist
                    foreach
                      my $album ( sort keys %{ $discs{$path}{$albumartist} } )
                    {    # sort by album
                        if (DEBUG) {
                            if ( is_os_type('Windows') ) {
                                printW( "Album: ", $album, "\n" );
                            }
                            else {
                                print "Album: ", $album, "\n";
                            }
                        }

                        my $filename;
                        if ( defined $albumartist && length($albumartist) > 0 )
                        {
                            if ( defined $album && length($album) > 0 ) {
                                $filename = join( '',
                                    $albumartist, ' - ', $album, '.m3u' );
                            }
                            else {
                                $filename = join( '', $albumartist, '.m3u' );
                            }
                        }
                        else {
                            if ( defined $album && length($album) > 0 ) {
                                $filename = join( '', $album, '.m3u' );
                            }
                            else {
                                $filename = join( '', basename($path), '.m3u' );
                            }
                        }
                        $filename = validateFilename($filename);
                        if (DEBUG) {
                            if ( is_os_type('Windows') ) {
                                printW( "Playlist's file name: ",
                                    $filename, "\n" );
                            }
                            else {
                                print "Playlist's file name: ", $filename, "\n";
                            }
                        }

                        my $fh;
                        my $outputpath;
                        if ( $OPT{'m3udir'} ) {
                            $outputpath = $OPT{'m3udir'};
                        }
                        else {
                            $outputpath = $path;
                        }
                        if (DEBUG) {
                            if ( is_os_type('Windows') ) {
                                printW( '$outputpath: ', $outputpath, "\n" );
                            }
                            else {
                                print '$outputpath: ', $outputpath, "\n";
                            }
                        }

                        if ( is_os_type('Windows') ) {
                            openL( \$fh, '>:encoding(UTF-8)',
                                catfile( $outputpath, $filename ) )
                              or die( "Cannot open file ",
                                $filename, ' in ', $outputpath, ': ', $^E,
                                ".\n" );
                        }
                        else {
                            open( $fh, '>', catfile( $outputpath, $filename ) )
                              or die( "Cannot open file ",
                                $filename, ' in ', $outputpath, ': ', $!,
                                ".\n" );
                        }

                        print $fh ("#EXTM3U\n");    # print the extended header

                        foreach my $disc ( sort { $a <=> $b }
                            keys %{ $discs{$path}{$albumartist}{$album} } )
                        {                           # sort by disc number
                            print "Disc (album): ", $disc, "\n" if DEBUG;
                            foreach my $track ( sort { $a <=> $b }
                                keys
                                %{ $discs{$path}{$albumartist}{$album}{$disc} }
                              )
                            {                       # sort by track number
                                print "  Track (album): ", $track, "\n"
                                  if DEBUG;
                                foreach my $file (
                                    sort { $a cmp $b } keys %{
                                        $discs{$path}{$albumartist}{$album}
                                          {$disc}{$track}
                                    }
                                  )
                                { # sort by file name in case of missing track number
                                    if (DEBUG) {
                                        if ( is_os_type('Windows') ) {
                                            printW( "    File (album): ",
                                                $file, "\n" );
                                        }
                                        else {
                                            print "    File (album): ", $file,
                                              "\n";
                                        }
                                    }
                                    printFile(
                                        $fh,
                                        $discs{$path}{$albumartist}{$album}
                                          {$disc}{$track}{$file}[0],
                                        $discs{$path}{$albumartist}{$album}
                                          {$disc}{$track}{$file}[1][0],
                                        $discs{$path}{$albumartist}{$album}
                                          {$disc}{$track}{$file}[1][1],
                                        $discs{$path}{$albumartist}{$album}
                                          {$disc}{$track}{$file}[1][4]
                                    );
                                }
                            }
                        }

                        close $fh;

                        if ( is_os_type('Windows') ) {
                            printW(
                                "Playlist '",
                                $filename,
                                "' for all artists and discs of the album '",
                                $album,
                                "' generated in ",
                                $outputpath,
                                "\n"
                            );
                        }
                        else {
                            print "Playlist '", $filename,
                              "' for all artists and discs of the album '",
                              $album, "' generated in ", $outputpath, "\n";
                        }
                    }
                }
            }
            else {
# %discs has the following keys: $discs{$path}{$artist}{$album}{$disc}{$track}{$$file[0]}
                foreach my $artist ( sort keys %{ $discs{$path} } )
                {    # sort by artist
                    if (DEBUG) {
                        if ( is_os_type('Windows') ) {
                            printW( "Artist: ", $artist, "\n" );
                        }
                        else {
                            print "Artist: ", $artist, "\n";
                        }
                    }

                    foreach my $album ( sort keys %{ $discs{$path}{$artist} } )
                    {    # sort by album
                        if (DEBUG) {
                            if ( is_os_type('Windows') ) {
                                printW( "Album: ", $album, "\n" );
                            }
                            else {
                                print "Album: ", $album, "\n";
                            }
                        }

                        my $filename;
                        if ( defined $artist && length($artist) > 0 ) {
                            if ( defined $album && length($album) > 0 ) {
                                $filename =
                                  join( '', $artist, ' - ', $album, '.m3u' );
                            }
                            else {
                                $filename = join( '', $artist, '.m3u' );
                            }
                        }
                        else {
                            if ( defined $album && length($album) > 0 ) {
                                $filename = join( '', $album, '.m3u' );
                            }
                            else {
                                $filename = join( '', basename($path), '.m3u' );
                            }
                        }
                        $filename = validateFilename($filename);
                        if (DEBUG) {
                            if ( is_os_type('Windows') ) {
                                printW( "Playlist's file name: ",
                                    $filename, "\n" );
                            }
                            else {
                                print "Playlist's file name: ", $filename, "\n";
                            }
                        }

                        my $fh;
                        my $outputpath;
                        if ( $OPT{'m3udir'} ) {
                            $outputpath = $OPT{'m3udir'};
                        }
                        else {
                            $outputpath = $path;
                        }
                        if (DEBUG) {
                            if ( is_os_type('Windows') ) {
                                printW( '$outputpath: ', $outputpath, "\n" );
                            }
                            else {
                                print '$outputpath: ', $outputpath, "\n";
                            }
                        }

                        if ( is_os_type('Windows') ) {
                            openL( \$fh, '>:encoding(UTF-8)',
                                catfile( $outputpath, $filename ) )
                              or die( "Cannot open file ",
                                $filename, ' in ', $outputpath, ': ', $^E,
                                ".\n" );
                        }
                        else {
                            open( $fh, '>', catfile( $outputpath, $filename ) )
                              or die( "Cannot open file ",
                                $filename, ' in ', $outputpath, ': ', $!,
                                ".\n" );
                        }

                        print $fh ("#EXTM3U\n");    # print the extended header

                        foreach my $disc ( sort { $a <=> $b }
                            keys %{ $discs{$path}{$artist}{$album} } )
                        {                           # sort by disc number
                            print "Disc: ", $disc, "\n" if DEBUG;
                            foreach my $track ( sort { $a <=> $b }
                                keys %{ $discs{$path}{$artist}{$album}{$disc} }
                              )
                            {                       # sort by track number
                                print "  Track: ", $track, "\n" if DEBUG;
                                foreach my $file (
                                    sort { $a cmp $b } keys %{
                                        $discs{$path}{$artist}{$album}{$disc}
                                          {$track}
                                    }
                                  )
                                { # sort by file name in case of missing track number
                                    if (DEBUG) {
                                        if ( is_os_type('Windows') ) {
                                            printW( "    File: ", $file, "\n" );
                                        }
                                        else {
                                            print "    File: ", $file, "\n";
                                        }
                                    }
                                    printFile(
                                        $fh,
                                        $discs{$path}{$artist}{$album}{$disc}
                                          {$track}{$file}[0],
                                        $discs{$path}{$artist}{$album}{$disc}
                                          {$track}{$file}[1][0],
                                        $discs{$path}{$artist}{$album}{$disc}
                                          {$track}{$file}[1][1],
                                        $discs{$path}{$artist}{$album}{$disc}
                                          {$track}{$file}[1][4]
                                    );
                                }
                            }
                        }

                        close $fh;

                        if ( is_os_type('Windows') ) {
                            printW(
                                "Playlist '",                     $filename,
                                "' for all discs of the album '", $album,
                                "' generated in ",                $outputpath,
                                "\n"
                            );
                        }
                        else {
                            print "Playlist '", $filename,
                              "' for all discs of the album '", $album,
                              "' generated in ", $outputpath, "\n";
                        }
                    }
                }
            }
        }
    }

    if ( $OPT{'disc'} ) {

        # Create a playlist for each disc of an album
        print
"\n---------------------------------\nGenerating playlists per album and per disc with all its tracks:\n\n";
        foreach my $path ( keys %discs ) {
            if ( $OPT{'album'} ) {

# %discs has the following keys: $discs{$path}{$albumartist}{$album}{$disc}{$track}{$$file[0]}

                foreach my $albumartist ( sort keys %{ $discs{$path} } )
                {    # sort by albumartist
                    foreach
                      my $album ( sort keys %{ $discs{$path}{$albumartist} } )
                    {    # sort by album
                        if (DEBUG) {
                            if ( is_os_type('Windows') ) {
                                printW( "Album: ", $album, "\n" );
                            }
                            else {
                                print "Album: ", $album, "\n";
                            }
                        }

                        my @discs =
                          keys %{ $discs{$path}{$albumartist}{$album} }
                          ;    # Array with all disc numbers of this album

                        foreach my $disc (@discs) {
                            print "Disc (album): ", $disc, "\n" if DEBUG;
                            my $cd = join( '', ' (CD ', $disc, ')' );
                            my $filename;
                            if ( defined $albumartist
                                && length($albumartist) > 0 )
                            {
                                if ( defined $album && length($album) > 0 ) {
                                    $filename = join( '',
                                        $albumartist, ' - ', $album, $cd,
                                        '.m3u' );
                                }
                                else {
                                    $filename =
                                      join( '', $albumartist, $cd, '.m3u' );
                                }
                            }
                            else {
                                if ( defined $album && length($album) > 0 ) {
                                    $filename = join( '', $album, $cd, '.m3u' );
                                }
                                else {
                                    $filename =
                                      join( '', basename($path), $cd, '.m3u' );
                                }
                            }
                            $filename = validateFilename($filename);
                            if (DEBUG) {
                                if ( is_os_type('Windows') ) {
                                    printW( "Playlist's file name: ",
                                        $filename, "\n" );
                                }
                                else {
                                    print "Playlist's file name: ", $filename,
                                      "\n";
                                }
                            }

                            my $fh;
                            my $outputpath;
                            if ( $OPT{'m3udir'} ) {
                                $outputpath = $OPT{'m3udir'};
                            }
                            else {
                                $outputpath = $path;
                            }
                            if (DEBUG) {
                                if ( is_os_type('Windows') ) {
                                    printW( '$outputpath: ', $outputpath,
                                        "\n" );
                                }
                                else {
                                    print '$outputpath: ', $outputpath, "\n";
                                }
                            }

                            if ( is_os_type('Windows') ) {
                                openL( \$fh, '>:encoding(UTF-8)',
                                    catfile( $outputpath, $filename ) )
                                  or die( "Cannot open file ",
                                    $filename, ' in ', $outputpath, ': ', $^E,
                                    ".\n" );
                            }
                            else {
                                open( $fh, '>',
                                    catfile( $outputpath, $filename ) )
                                  or die( "Cannot open file ",
                                    $filename, ' in ', $outputpath, ': ', $!,
                                    ".\n" );
                            }

                            print $fh ("#EXTM3U\n"); # print the extended header

                            foreach my $track ( sort { $a <=> $b }
                                keys
                                %{ $discs{$path}{$albumartist}{$album}{$disc} }
                              )
                            {                        # sort by track number
                                print "  Track (album): ", $track, "\n"
                                  if DEBUG;
                                foreach my $file (
                                    sort { $a cmp $b } keys %{
                                        $discs{$path}{$albumartist}{$album}
                                          {$disc}{$track}
                                    }
                                  )
                                { # sort by file name in case of missing track number
                                    if (DEBUG) {
                                        if ( is_os_type('Windows') ) {
                                            printW( "    File (album): ",
                                                $file, "\n" );
                                        }
                                        else {
                                            print "    File (album): ", $file,
                                              "\n";
                                        }
                                    }
                                    printFile(
                                        $fh,
                                        $discs{$path}{$albumartist}{$album}
                                          {$disc}{$track}{$file}[0],
                                        $discs{$path}{$albumartist}{$album}
                                          {$disc}{$track}{$file}[1][0],
                                        $discs{$path}{$albumartist}{$album}
                                          {$disc}{$track}{$file}[1][1],
                                        $discs{$path}{$albumartist}{$album}
                                          {$disc}{$track}{$file}[1][4]
                                    );
                                }
                            }

                            close $fh;

                            if ( is_os_type('Windows') ) {
                                printW(
                                    "Playlist '",            $filename,
                                    "' generated for disc ", $disc,
                                    " of the album '",       $album,
                                    "' in ",                 $outputpath,
                                    "\n"
                                );
                            }
                            else {
                                print "Playlist '", $filename,
                                  "' generated for disc ", $disc,
                                  " of the album '", $album, "' in ",
                                  $outputpath, "\n";
                            }
                        }
                    }
                }
            }
            else {
# %discs has the following keys: $discs{$path}{$artist}{$album}{$disc}{$track}{$$file[0]}
                foreach my $artist ( sort keys %{ $discs{$path} } )
                {    # sort by artist
                    if (DEBUG) {
                        if ( is_os_type('Windows') ) {
                            printW( "Artist: ", $artist, "\n" );
                        }
                        else {
                            print "Artist: ", $artist, "\n";
                        }
                    }

                    foreach my $album ( sort keys %{ $discs{$path}{$artist} } )
                    {    # sort by album
                        if (DEBUG) {
                            if ( is_os_type('Windows') ) {
                                printW( "Album: ", $album, "\n" );
                            }
                            else {
                                print "Album: ", $album, "\n";
                            }
                        }

                        my @discs = keys %{ $discs{$path}{$artist}{$album} }
                          ;    # Array with all disc numbers of this album

                        foreach my $disc (@discs) {
                            print "Disc: ", $disc, "\n" if DEBUG;
                            my $cd = join( '', ' (CD ', $disc, ')' );
                            my $filename;
                            if ( defined $artist && length($artist) > 0 ) {
                                if ( defined $album && length($album) > 0 ) {
                                    $filename = join( '',
                                        $artist, ' - ', $album, $cd, '.m3u' );
                                }
                                else {
                                    $filename =
                                      join( '', $artist, $cd, '.m3u' );
                                }
                            }
                            else {
                                if ( defined $album && length($album) > 0 ) {
                                    $filename = join( '', $album, $cd, '.m3u' );
                                }
                                else {
                                    $filename =
                                      join( '', basename($path), $cd, '.m3u' );
                                }
                            }
                            $filename = validateFilename($filename);
                            if (DEBUG) {
                                if ( is_os_type('Windows') ) {
                                    printW( "Playlist's file name: ",
                                        $filename, "\n" );
                                }
                                else {
                                    print "Playlist's file name: ", $filename,
                                      "\n";
                                }
                            }

                            my $fh;
                            my $outputpath;
                            if ( $OPT{'m3udir'} ) {
                                $outputpath = $OPT{'m3udir'};
                            }
                            else {
                                $outputpath = $path;
                            }
                            if (DEBUG) {
                                if ( is_os_type('Windows') ) {
                                    printW( '$outputpath: ', $outputpath,
                                        "\n" );
                                }
                                else {
                                    print '$outputpath: ', $outputpath, "\n";
                                }
                            }

                            if ( is_os_type('Windows') ) {
                                openL( \$fh, '>:encoding(UTF-8)',
                                    catfile( $outputpath, $filename ) )
                                  or die( "Cannot open file ",
                                    $filename, ' in ', $outputpath, ': ', $^E,
                                    ".\n" );
                            }
                            else {
                                open( $fh, '>',
                                    catfile( $outputpath, $filename ) )
                                  or die( "Cannot open file ",
                                    $filename, ' in ', $outputpath, ': ', $!,
                                    ".\n" );
                            }

                            print $fh ("#EXTM3U\n"); # print the extended header

                            foreach my $track ( sort { $a <=> $b }
                                keys %{ $discs{$path}{$artist}{$album}{$disc} }
                              )
                            {                        # sort by track number
                                print "  Track: ", $track, "\n" if DEBUG;
                                foreach my $file (
                                    sort { $a cmp $b } keys %{
                                        $discs{$path}{$artist}{$album}{$disc}
                                          {$track}
                                    }
                                  )
                                { # sort by file name in case of missing track number
                                    if (DEBUG) {
                                        if ( is_os_type('Windows') ) {
                                            printW( "    File: ", $file, "\n" );
                                        }
                                        else {
                                            print "    File: ", $file, "\n";
                                        }
                                    }
                                    printFile(
                                        $fh,
                                        $discs{$path}{$artist}{$album}{$disc}
                                          {$track}{$file}[0],
                                        $discs{$path}{$artist}{$album}{$disc}
                                          {$track}{$file}[1][0],
                                        $discs{$path}{$artist}{$album}{$disc}
                                          {$track}{$file}[1][1],
                                        $discs{$path}{$artist}{$album}{$disc}
                                          {$track}{$file}[1][4]
                                    );
                                }
                            }

                            close $fh;

                            if ( is_os_type('Windows') ) {
                                printW(
                                    "Playlist '",            $filename,
                                    "' generated for disc ", $disc,
                                    " of the album '",       $album,
                                    "' in ",                 $outputpath,
                                    "\n"
                                );
                            }
                            else {
                                print "Playlist '", $filename,
                                  "' generated for disc ", $disc,
                                  " of the album '", $album, "' in ",
                                  $outputpath, "\n";
                            }
                        }
                    }
                }
            }
        }

   # Playlists for each disc of an album with one album per directory generated.
    }
}

############################################################################
# Convert an array to a string if the input is an array
sub arrayToString($$) {
    my $input     = $_[0];
    my $separator = $_[1];
    if (    defined($input)
        and length($input) > 0
        and defined( reftype($input) )
        and reftype($input) eq 'ARRAY' )
    {
        if ( defined($separator) and length($separator) > 0 ) {
            return join( $separator, @$input );
        }
        else {
            return $$input[0];
        }
    }
    else {
        return $input;
    }
}

##############################################################################
# fisher_yates_shuffle( \@array ) : generate a random permutation
# of @array in place
sub fisher_yates_shuffle {
    my $array = shift;
    my $i;
    for ( $i = @$array ; --$i ; ) {
        my $j = int rand( $i + 1 );
        next if $i == $j;
        @$array[ $i, $j ] = @$array[ $j, $i ];
    }
}

############################################################################
# prints a short help text
sub help() {
    my $audio = $LIST_AUDIO ? 'installed' : 'NOT installed';
    my $flac  = $LIST_FLAC  ? 'installed' : 'NOT installed';
    my $mp3   = $LIST_MP3   ? 'installed' : 'NOT installed';
    my $ogg   = $LIST_OGG   ? 'installed' : 'NOT installed';

    print <<STOP;
M3U Playlist Director, Version $version

Usage: m3upd.pl [-b] [-a] [-d] [-h] [-i] [-m=<directory>] [-r] [-t] <audio-dir[s]>

-album       Create playlists for albums that may have more than one artist.
or           This option can only be used in combination with the options
-b           'all' and 'disc'. All albums' audio files must be stored in the
             same directory. Audio files from different albums may be stored
             in the same directory.  

-all         Create playlists for all audio files of all discs for each
or           album. An album may have one or more discs. An album's and all
-a           its discs' audio files must be stored in the same directory.

-disc        Create a playlist per disc for albums with more than one disc.
or           An album's and all its discs' audio files must be stored in the
-d           same directory.

-help        Shows this help page.
or
-h

-interactive The options and directories are set by interaction with the
or           user. The options provided via the command line are used as
-i           default values.

-m3udir      All M3U playlists are stored in this given directory. Otherwise,
or           the playlists are stored in the same directories like their
-m           referenced audio files.

-random      Randomize playlist order if the output is printed to STDOUT.
or           This option is ignored if the options 'all' or 'disc' are used.
-r

-twonky      Strange relative paths for the Twonky (tm) Server, like
or           ../artist - album/artist - album (CD 1) - 01 title.flac
-t           for a playlist stored in
             /home/user/Music/artist - album
             
<audio-dir>  Search this directory recursivly for audio files.

This Perl script generates extended M3U playlists from a given directory
for many media players.
The playlists are printed to files encoded in UTF-8 or to STDOUT.
The path references to the media files in the playlists are relative to the
playlists' directories. Therefore, the playlists and the media files cannot
be stored on different drives in DOS and Windows OS systems.
Extended M3U files contain additional information like length of the
song and information from ID3 tags, for instance.

If you provide a '-' instead of an <audio-dir> directory, the script will
read the audio files' directories from STDIN.

The playlists' directory (i.e. option m3udir) and the audio-dir(s) provided
via the command line do not support non-ANSI encoded characters on DOS and
Windows OS systems because of Perl's limitations on these operating systems.

For running this script, you need the following Perl modules:

Audio::Scan                     for many media files ($audio)
or
Audio::FLAC::Header             for FLAC files       ($flac)
MP3::Info                       for MP3 files        ($mp3)
Ogg::Vorbis::Header::PurePerl   for OGG Vorbis files ($ogg)

The script autodetects the installed modules. If a module is not installed,
the corresponding file type is ignored.
Audio::Scan is the preferred module because it supports many media file
types.
_____________________________________________________________________________
m3upd.pl - Generates extended M3U playlists
Based on extm3u.pl (2011-12-01) by Andreas Gohr <andi\@splitbrain.org>
Copyright (C) 2022 Rainer Hillebrand <da35ph6o13ps7ffaivve\@webcab.de>

This program is free software. You can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.
See COPYING for details
STOP
}

##############################################################################
# Print extended file info
sub printFile($$$$$) {
    my $fh     = $_[0];
    my $file   = $_[1];
    my $sec    = $_[2];
    my $artist = $_[3];
    my $title  = $_[4];

    my ( $filenamewithoutsuffix, $path, $suffix ) =
      fileparse( $file, '\.[\w\d]+' );
    my $filename = join( '', $filenamewithoutsuffix, $suffix );

    if ( !defined($sec) || length($sec) < 1 ) {
        $sec = '';
    }

    if (   defined $artist
        && length($artist) > 0
        && defined $title
        && length($title) > 0 )
    {
        if ( $OPT{'all'} || $OPT{'disc'} ) {
            print $fh ( "#EXTINF:", $sec, ",", $artist, " - ", $title, "\n" );
        }
        else {
            if ( is_os_type('Windows') ) {
                printW( "#EXTINF:", $sec, ",", $artist, " - ", $title, "\n" );
            }
            else {
                print( "#EXTINF:", $sec, ",", $artist, " - ", $title, "\n" );
            }
        }
    }
    else {
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( '    File name without extension (missing title): ',
                    $filenamewithoutsuffix, "\n" );
            }
            else {
                print( '    File name without extension (missing title): ',
                    $filenamewithoutsuffix, "\n" );
            }
        }

        if ( $OPT{'all'} || $OPT{'disc'} ) {
            print $fh ( "#EXTINF:", $sec, ",", $filenamewithoutsuffix, "\n" );
        }
        else {
            if ( is_os_type('Windows') ) {
                printW( "#EXTINF:", $sec, ",", $filenamewithoutsuffix, "\n" );
            }
            else {
                print( "#EXTINF:", $sec, ",", $filenamewithoutsuffix, "\n" );
            }
        }
    }

    if ( $OPT{'all'} || $OPT{'disc'} ) {
        if ( $OPT{'m3udir'} ) {
            my $relative_path =
              File::Spec::Unix->abs2rel( $path, $OPT{'m3udir'} );
            if (DEBUG) {
                if ( is_os_type('Windows') ) {
                    printW( '$path:          ', $path,          "\n" );
                    printW( '$relative_path: ', $relative_path, "\n" );
                    printW( '$filename:      ', $filename,      "\n" );
                    printW(
                        'File::Spec::Unix->catfile($relative_path,$filename): ',
                        File::Spec::Unix->catfile( $relative_path, $filename ),
                        "\n"
                    );
                }
                else {
                    print '$path:          ', $path,          "\n";
                    print '$relative_path: ', $relative_path, "\n";
                    print '$filename:      ', $filename,      "\n";
                    print
                      'File::Spec::Unix->catfile($relative_path,$filename): ',
                      File::Spec::Unix->catfile( $relative_path, $filename ),
                      "\n";
                }
            }
            print $fh (
                File::Spec::Unix->catfile( $relative_path, $filename ), "\n"
            );
        }
        elsif ( $OPT{'twonky'} ) {
            my $parent = basename($path);
            if (DEBUG) {
                if ( is_os_type('Windows') ) {
                    printW( '$path: ',   $path,   "\n" );
                    printW( '$parent: ', $parent, "\n" );
                    printW(
                        'File::Spec::Unix->catfile(' .. ',$parent,$filename): ',
                        File::Spec::Unix->catfile( '..', $parent, $filename ),
                        "\n"
                    );
                }
                else {
                    print '$path: ',   $path,   "\n";
                    print '$parent: ', $parent, "\n";
                    print 'File::Spec::Unix->catfile(' ..
                      ',$parent,$filename): ',
                      File::Spec::Unix->catfile( '..', $parent, $filename ),
                      "\n";
                }
            }
            print $fh (
                File::Spec::Unix->catfile( '..', $parent, $filename ), "\n"
            );
        }
        else {
            print $fh ( $filename, "\n" );
        }
    }
    else {
        if ( is_os_type('Windows') ) {
            printW( $file, "\n" );
        }
        else {
            print( $file, "\n" );
        }
    }
}

##############################################################################
# Recursively scan a directory
sub readdirectory($) {

    # Open directory and read contents
    my $path = shift;
    print( 'Search for files in this directory: ', $path, "\n" ) if DEBUG;

    opendir( my $directory, $path ) or die "Can't opendir() ", $path, ': ', $!;
    my @files =
      map { File::Spec::Unix->catfile( $path, decode( 'UTF-8', $_ ) ) }
      grep { $_ !~ /^\./ } readdir($directory);
    closedir($directory);

    my @collectedfiles;

    foreach my $file ( sort(@files) ) {
        print( 'Is this a directory or a file? ', $file, "\n" ) if DEBUG;
        if ( -d $file ) {

            # recursion
            if ( $file ne '-' and $file ne '..' and $file ne '.' ) {
                print "Directory found.\n" if DEBUG;
                push( @collectedfiles, readdirectory($file) );
            }
        }
        elsif ( -f $file ) {
            print "File found.\n" if DEBUG;
            push( @collectedfiles, $file );
        }

    }
    return @collectedfiles;
}

##############################################################################
# Recursively scan a directory on Windows OS
sub readdirectoryWindows($) {

    # Open directory and read contents
    my $path = shift;
    printW( 'Search for files in this directory: ', $path, "\n" ) if DEBUG;

    my $dir = Win32::LongPath->new();
    $dir->opendirL($path) or die("Unable to open $path ($^E)\n");

    my @collectedfiles;
    foreach my $file ( $dir->readdirL() ) {
        if ( $file eq '..' or $file eq '.' ) {

            # skip parent and current directory
            next;
        }

        printW( '$file: ', $file, "\n" ) if DEBUG;

        # get file stats
        my $name = File::Spec::Unix->catfile( $path, $file );
        my $stat = lstatL($name) or die "Unable to stat $name ($^E)";
        printW( '$stat->{attribs}: ', $stat->{attribs}, "\n" ) if DEBUG;

# recurse if directory
# See https://docs.microsoft.com/en-us/windows/win32/fileio/file-attribute-constants for Win32 file attributes.
# FILE_ATTRIBUTE_DIRECTORY:       16 (0x10)
# FILE_ATTRIBUTE_REPARSE_POINT: 1024 (0x400)
        if ( ( $stat->{attribs} & 16 ) or ( $stat->{attribs} & 1024 ) ) {
            printW( 'Directory found: ', $name, "\n" ) if DEBUG;
            push( @collectedfiles, readdirectoryWindows($name) );
            next;
        }
        else {
            printW( 'File found (maybe): ', $name, "\n" )
              if DEBUG;    # It is not a directory but may be a file.
            push( @collectedfiles, $name );
        }
    }
    $dir->closedirL();
    return @collectedfiles;
}

##############################################################################
# Read a given directory and its subdirectories
sub readFiles($$) {
    my $path    = $_[0];
    my $collect = $_[1];

    my @files;
    if ( $path eq '-' ) {
        @files = <STDIN>;
        chomp(@files);
    }
    else {
        if ( is_os_type('Windows') ) {
            push( @files, readdirectoryWindows($path) );
        }
        else {
            push( @files, readdirectory($path) );
        }
    }
    print '@files: ', "\n", Dumper(@files) if DEBUG;

    my @allfiles;

    if ( !$OPT{'all'} && !$OPT{'disc'} ) {
        print("#EXTM3U\n")
          ;    # print the M3U extended header to STDOUT, no files are created
    }

    foreach my $file ( sort(@files) ) {
        my $fileL = $file;
        if ( is_os_type('Windows') ) {

# Audio::FLAC::Header and other modules cannot access files in Windows with Unicode
# encoding or long file paths. Therefore, the short path of a file is used.
            $fileL = shortpathL($file);
            if ( length($fileL) < 1 ) {
                die "Unable to get shortpathL() for ", $file, "\n";
            }
        }

        # get audio file infos
        my $info = undef
          ; # $info will have the following elements: (0: sec, 1: artist, 2: album, 3: albumartist, 4: title, 5: tracknumber, 6: discnumber);
        if ($LIST_AUDIO) {
            if ( Audio::Scan->is_supported($fileL) ) {
                $info = scanAudio($fileL);
            }
            else {
                if (DEBUG) {
                    if ( is_os_type('Windows') ) {
                        printW( 'This file type is not supported: ',
                            $fileL, "\n" );
                    }
                    else {
                        print 'This file type is not supported: ', $fileL, "\n";
                    }
                }
            }
        }
        else {
            if ( $LIST_MP3 and $file =~ /\.mp3$/i ) {
                $info = scanMP3($fileL);
            }
            elsif ( $LIST_FLAC and $file =~ /\.flac?$/i ) {
                $info = scanFLAC($fileL);
            }
            elsif ( $LIST_OGG and $file =~ /\.ogg$/i ) {
                $info = scanOGG($fileL);
            }
            elsif ( $file =~ /\.(mp3|flac?|ogg|wav)$/i )
            {    # audio file, but there is no module to process it
                $info = undef;
            }
        }

        # output file
        if ($info) {
            if ($collect) {
                push( @allfiles, [ $file, $info ] );
            }
            else {
                printFile( undef, $file, $$info[0], $$info[1], $$info[4] );
            }
        }
    }
    return @allfiles;
}

###############################################################################
# Retrieve some data from an audio or video file
sub scanAudio($) {
    my $file = $_[0];
    local $ENV{AUDIO_SCAN_NO_ARTWORK} = 1;
    my $s = Audio::Scan->scan($file);

    print "Media file: ", $file, "\n" if DEBUG;
    print "All info and tags: ", Dumper( \$s ) if DEBUG;

    my $tags = $s->{tags};

    my $artist =
         $tags->{ARTIST}
      || $tags->{TPE1}
      || $tags->{ART}
      || $tags->{Author}
      || $tags->{IART};
    if ( defined($artist) ) {
        $artist = arrayToString( $artist, ', ' );
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Artist: ", $artist, "\n" );
            }
            else {
                print "Artist: ", $artist, "\n";
            }
        }
    }

    my $album =
         $tags->{ALBUM}
      || $tags->{ALB}
      || $tags->{AlbumTitle}
      || $tags->{IPRD}
      || $tags->{TALB}
      || $tags->{'WM/AlbumTitle'};
    if ( defined($album) ) {
        $album = arrayToString( $album, ', ' );
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Album: ", $album, "\n" );
            }
            else {
                print "Album: ", $album, "\n";
            }
        }
    }

    my $albumartist =
         $tags->{ALBUMARTIST}
      || $tags->{TPE2}
      || $tags->{AART}
      || $tags->{ART}
      || $tags->{AlbumArtist}
      || $tags->{'WM/ALBUMARTIST'};
    if ( defined($albumartist) ) {
        $albumartist = arrayToString( $albumartist, ', ' );
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Album artist: ", $albumartist, "\n" );
            }
            else {
                print "Album artist: ", $albumartist, "\n";
            }
        }
    }

    my $title =
         $tags->{TITLE}
      || $tags->{TIT2}
      || $tags->{NAM}
      || $tags->{Title}
      || $tags->{INAM};
    if ( defined($title) ) {
        $title = arrayToString( $title, ', ' );
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Title: ", $title, "\n" );
            }
            else {
                print "Title: ", $title, "\n";
            }
        }
    }

    my $tracknumber =
         $tags->{TRACKNUMBER}
      || $tags->{TRCK}
      || $tags->{TRKN}
      || $tags->{TrackNumber}
      || $tags->{ITRK}
      || $tags->{TRACK}
      || $tags->{'WM/TrackNumber'};
    if ( defined($tracknumber) ) {
        $tracknumber = arrayToString( $tracknumber, '' );
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Track number: ", $tracknumber, "\n" );
            }
            else {
                print "Track number: ", $tracknumber, "\n";
            }
        }
    }

    my $discnumber =
         $tags->{DISCNUMBER}
      || $tags->{TPOS}
      || $tags->{DISK}
      || $tags->{PartOfSet}
      || $tags->{'WM/PartOfSet'};
    if ( defined($discnumber) ) {
        $discnumber = arrayToString( $discnumber, '' );
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Disc number: ", $discnumber, "\n" );
            }
            else {
                print "Disc number: ", $discnumber, "\n";
            }
        }
    }

    my $sec = $s->{info}->{song_length_ms};
    if ( defined($sec) ) {
        $sec = int($sec / 1000);
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Seconds: ", $sec, "\n\n" );
            }
            else {
                print "Seconds: ", $sec, "\n\n";
            }
        }
    }

    return [
        $sec,   $artist,      $album, $albumartist,
        $title, $tracknumber, $discnumber
    ];
}

###############################################################################
# Retrieve some data from a FLAC file
sub scanFLAC($) {
    my $file = $_[0];
    my $flac = new Audio::FLAC::Header($file);

    my $tags = $flac->tags();

    print "FLAC file: ", $file, "\n" if DEBUG;
    print "FLAC comment fields:\nAll comment fields: ", Dumper( \%$tags )
      if DEBUG;

    my %tagslc
      ; # will store all tags with lower case field names because the tags may be set with lower case letters
    foreach ( keys %$tags ) {
        $tagslc{ lc($_) } = $_
          ; # The tag name (i.e. field name) is case-insensitive according to the
         # "Ogg Vorbis I format specification: comment field and header specification"
         # (see https://www.xiph.org/vorbis/doc/v-comment.html).
    }

    my ( $artist, $album, $albumartist, $title, $tracknumber, $discnumber );
    if ( defined $tagslc{'artist'} ) {
        $artist = $flac->tags( $tagslc{'artist'} );
        if ( defined $artist ) {
            $artist = arrayToString( $artist, ', ' );
            $artist = decode( 'UTF-8', $artist )
              ; # Transcode this comment field from UTF-8 to the Perl internal utf8 encoding.
        }
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Artist: ", $artist, "\n" );
            }
            else {
                print "Artist: ", $artist, "\n";
            }
        }
    }
    if ( defined $tagslc{'album'} ) {
        $album = $flac->tags( $tagslc{'album'} );
        if ( defined $album ) {
            $album = arrayToString( $album, ', ' );
            $album = decode( 'UTF-8', $album )
              ; # Transcode this comment field from UTF-8 to the Perl internal utf8 encoding.
        }
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Album: ", $album, "\n" );
            }
            else {
                print "Album: ", $album, "\n";
            }
        }
    }
    if ( defined $tagslc{'albumartist'} ) {
        $albumartist = $flac->tags( $tagslc{'albumartist'} );
        if ( defined $albumartist ) {
            $albumartist = arrayToString( $albumartist, ', ' );
            $albumartist = decode( 'UTF-8', $albumartist )
              ; # Transcode this comment field from UTF-8 to the Perl internal utf8 encoding.
        }
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Album artist: ", $albumartist, "\n" );
            }
            else {
                print "Album artist: ", $albumartist, "\n";
            }
        }
    }
    if ( defined $tagslc{'title'} ) {
        $title = $flac->tags( $tagslc{'title'} );
        if ( defined $title ) {
            $title = arrayToString( $title, ', ' );
            $title = decode( 'UTF-8', $title )
              ; # Transcode this comment field from UTF-8 to the Perl internal utf8 encoding.
        }
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Title: ", $title, "\n" );
            }
            else {
                print "Title: ", $title, "\n";
            }
        }
    }
    if ( defined $tagslc{'tracknumber'} ) {
        $tracknumber = $flac->tags( $tagslc{'tracknumber'} );
        if ( defined $tracknumber ) {
            $tracknumber = arrayToString( $tracknumber, '' );
            $tracknumber = decode( 'UTF-8', $tracknumber )
              ; # Transcode this comment field from UTF-8 to the Perl internal utf8 encoding.
        }
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Track number: ", $tracknumber, "\n" );
            }
            else {
                print "Track number: ", $tracknumber, "\n";
            }
        }
    }
    if ( defined $tagslc{'discnumber'} ) {
        $discnumber = $flac->tags( $tagslc{'discnumber'} );
        if ( defined $discnumber ) {
            $discnumber = arrayToString( $discnumber, '' );
            $discnumber = decode( 'UTF-8', $discnumber )
              ; # Transcode this comment field from UTF-8 to the Perl internal utf8 encoding.
        }
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Disc number: ", $discnumber, "\n" );
            }
            else {
                print "Disc number: ", $discnumber, "\n";
            }
        }
    }
    my $sec = int( $flac->{trackTotalLengthSeconds} );
    if (DEBUG) {
        if ( is_os_type('Windows') ) {
            printW( "Seconds: ", $sec, "\n\n" );
        }
        else {
            print "Seconds: ", $sec, "\n\n";
        }
    }

    return [
        $sec,   $artist,      $album, $albumartist,
        $title, $tracknumber, $discnumber
    ];
}

###############################################################################
# Retrieve some data from an MP3 file
sub scanMP3($) {
    my $file = $_[0];

    my $info = MP3::Info::get_mp3info($file);
    my $tags = MP3::Info::get_mp3tag( $file, undef, 2 );

    my $sec = $info->{SECS};
    if ( defined($sec) and length($sec) > 0 ) {
        $sec = int($sec);
    }
    if (DEBUG) {
        if ( is_os_type('Windows') ) {
            printW( "Seconds: ", $sec, "\n" );
        }
        else {
            print "Seconds: ", $sec, "\n";
        }
    }

    print "MP3 file: ", $file, "\n" if DEBUG;
    print "MP3 information:\n" if DEBUG;
    print '$info: ',        Dumper( \%$info ) if DEBUG;
    print "All MP3 tags: ", Dumper( \%$tags ) if DEBUG;

    print "Required MP3 tags:\n" if DEBUG;
    my $artist = $tags->{TPE1} || $tags->{ARTIST};
    if ( defined $artist ) {
        $artist = arrayToString( $artist, ', ' );
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Artist: ", $artist, "\n" );
            }
            else {
                print "Artist: ", $artist, "\n";
            }
        }
    }
    my $album = $tags->{TALB} || $tags->{ALBUM};
    if ( defined $album ) {
        $album = arrayToString( $album, ', ' );
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Album: ", $album, "\n" );
            }
            else {
                print "Album: ", $album, "\n";
            }
        }
    }
    my $albumartist = $tags->{TPE2};
    if ( defined $albumartist ) {
        $albumartist = arrayToString( $albumartist, ', ' );
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Album artist: ", $albumartist, "\n" );
            }
            else {
                print "Album artist: ", $albumartist, "\n";
            }
        }
    }
    my $title = $tags->{TIT2} || $tags->{TITLE};
    if ( defined $title ) {
        $title = arrayToString( $title, ', ' );
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Title: ", $title, "\n" );
            }
            else {
                print "Title: ", $title, "\n";
            }
        }
    }
    my $tracknumber = $tags->{TRCK} || $tags->{TRACKNUM};
    if ( defined $tracknumber ) {
        $tracknumber = arrayToString( $tracknumber, '' );
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Track number: ", $tracknumber, "\n" );
            }
            else {
                print "Track number: ", $tracknumber, "\n";
            }
        }
    }
    my $discnumber = $tags->{TPOS};
    if ( defined $discnumber ) {
        $discnumber = arrayToString( $discnumber, '' );
        if (DEBUG) {
            if ( is_os_type('Windows') ) {
                printW( "Disc number: ", $discnumber, "\n\n" );
            }
            else {
                print "Disc number: ", $discnumber, "\n\n";
            }
        }
    }

    return [
        $sec,   $artist,      $album, $albumartist,
        $title, $tracknumber, $discnumber
    ];
}

###############################################################################
# Retrieve some data from an OGG Vorbis file
sub scanOGG($) {
    my $file = $_[0];
    print "File: ", $file, "\n" if DEBUG;
    my $ogg = new Ogg::Vorbis::Header::PurePerl($file);
    return if ( !defined($ogg) );    # this is no ogg

    my $sec = int( $ogg->info->{'length'} );

    # This strange construction is to fetch artist and title regardless
    # of tag case
    my ( $artist, $album, $albumartist, $title, $tracknumber, $discnumber );
    my $tags = join( "\n", $ogg->comment_tags() );
    if ( $tags =~ m/^(artist)$/mi ) {
        my @array = $ogg->comment($1);
        if ( $#array >= 0 ) {
            $artist = arrayToString( \@array, ', ' );
            $artist = decode( 'UTF-8', $artist )
              ; # Transcode this comment field from UTF-8 to the Perl internal utf8 encoding.
            if (DEBUG) {
                if ( is_os_type('Windows') ) {
                    printW( "Artist: ", $artist, "\n" );
                }
                else {
                    print "Artist: ", $artist, "\n";
                }
            }
        }
    }
    if ( $tags =~ m/^(album)$/mi ) {
        my @array = $ogg->comment($1);
        if ( $#array >= 0 ) {
            $album = arrayToString( \@array, ', ' );
            $album = decode( 'UTF-8', $album )
              ; # Transcode this comment field from UTF-8 to the Perl internal utf8 encoding.
            if (DEBUG) {
                if ( is_os_type('Windows') ) {
                    printW( "Album: ", $album, "\n" );
                }
                else {
                    print "Album: ", $album, "\n";
                }
            }
        }
    }
    if ( $tags =~ m/^(albumartist)$/mi ) {
        my @array = $ogg->comment($1);
        if ( $#array >= 0 ) {
            $albumartist = arrayToString( \@array, ', ' );
            $albumartist = decode( 'UTF-8', $albumartist )
              ; # Transcode this comment field from UTF-8 to the Perl internal utf8 encoding.
            if (DEBUG) {
                if ( is_os_type('Windows') ) {
                    printW( "Album artist: ", $albumartist, "\n" );
                }
                else {
                    print "Album artist: ", $albumartist, "\n";
                }
            }
        }
    }
    if ( $tags =~ m/^(title)$/mi ) {
        my @array = $ogg->comment($1);
        if ( $#array >= 0 ) {
            $title = arrayToString( \@array, ', ' );
            $title = decode( 'UTF-8', $title )
              ; # Transcode this comment field from UTF-8 to the Perl internal utf8 encoding.
            if (DEBUG) {
                if ( is_os_type('Windows') ) {
                    printW( "Title: ", $title, "\n" );
                }
                else {
                    print "Title: ", $title, "\n";
                }
            }
        }
    }
    if ( $tags =~ m/^(tracknumber)$/mi ) {
        $tracknumber = ( $ogg->comment($1) )[0];
        if ( defined $tracknumber ) {
            if (DEBUG) {
                if ( is_os_type('Windows') ) {
                    printW( "Track number: ", $tracknumber, "\n" );
                }
                else {
                    print "Track number: ", $tracknumber, "\n";
                }
            }
        }
    }
    if ( $tags =~ m/^(discnumber)$/mi ) {
        $discnumber = ( $ogg->comment($1) )[0];
        if ( defined $discnumber ) {
            if (DEBUG) {
                if ( is_os_type('Windows') ) {
                    printW( "Disc number: ", $discnumber, "\n\n" );
                }
                else {
                    print "Disc number: ", $discnumber, "\n\n";
                }
            }
        }
    }

    return [
        $sec,   $artist,      $album, $albumartist,
        $title, $tracknumber, $discnumber
    ];
}

##############################################################################
# Replace characters from a file name by an underscore that shall not be used on Windows systems.
# (see https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file?redirectedfrom=MSDN#naming-conventions)
# Some whitespaces are deleted as well.
sub validateFilename($) {
    my $filename = shift;
    $filename =~ s#[\<\>\:\"\/\\\|\?\*]+#_#g;
    $filename =~ s#\s{2,}# #g;
    $filename =~ s#^\s+##g;
    $filename =~ s#\s+(\.[\w\d]+)$#$1#g;
    return $filename;
}

##############################################################################
# Print the version number of this Perl script
sub version() {
    print "M3U Playlist Director\n", 'Version: ', $version, "\n";
    exit -1;
}
